@@ -3,7 +3,10 @@ import os
 ###########################################################
 # Set version
 
-with open(os.path.join(os.path.abspath(os.path.dirname(__file__)), 'version'), encoding='utf-8') as v:
+with open(
+    os.path.join(os.path.abspath(os.path.dirname(__file__)), "version"),
+    encoding="utf-8",
+) as v:
     __version__ = v.read().rstrip()
 
 ###########################################################
