@@ -4,6 +4,7 @@ import ctypes
 # Some constants taken from cuda.h
 CUDA_SUCCESS = 0
 
+
 def get_gpu_number():
     """
     Return number of GPU by reading libcuda.
@@ -11,7 +12,7 @@ def get_gpu_number():
     From https://gist.github.com/f0k/0d6431e3faa60bffc788f8b4daa029b1
     credit: Jan Schlüter
     """
-    libnames = ('libcuda.so', 'libcuda.dylib', 'cuda.dll')
+    libnames = ("libcuda.so", "libcuda.dylib", "cuda.dll")
     for libname in libnames:
         try:
             cuda = ctypes.CDLL(libname)
@@ -30,14 +31,18 @@ def get_gpu_number():
     if result != CUDA_SUCCESS:
         # cuda.cuGetErrorString(result, ctypes.byref(error_str))
         # print("[pyKeOps]: cuInit failed with error code %d: %s" % (result, error_str.value.decode()))
-        print("[pyKeOps]: Warning, cuda was detected, but driver API could not be initialized. Switching to cpu only.")
+        print(
+            "[pyKeOps]: Warning, cuda was detected, but driver API could not be initialized. Switching to cpu only."
+        )
         return 0
 
     result = cuda.cuDeviceGetCount(ctypes.byref(nGpus))
     if result != CUDA_SUCCESS:
         # cuda.cuGetErrorString(result, ctypes.byref(error_str))
         # print("[pyKeOps]: cuDeviceGetCount failed with error code %d: %s" % (result, error_str.value.decode()))
-        print("[pyKeOps]: Warning, cuda was detected, driver API has been initialized, but no working GPU has been found. Switching to cpu only.")
+        print(
+            "[pyKeOps]: Warning, cuda was detected, driver API has been initialized, but no working GPU has been found. Switching to cpu only."
+        )
         return 0
 
-    return nGpus.value
\ No newline at end of file
+    return nGpus.value
