@@ -69,7 +69,8 @@ at::Tensor allocate_result_array_gpu< at::Tensor, __TYPE__ >(int* shape_out, int
   c10::ArrayRef < int64_t > shape_out_array(shape_out_long, (int64_t) nbatchdims + 2);
 
   // Create a new result array of shape [A, .., B, M, D] or [A, .., B, N, D]:
-  return torch::empty(shape_out_array, at::device({at::kCUDA, Device_Id}).dtype(AT_kTYPE).requires_grad(true));
+  return torch::empty(shape_out_array, at::device({at::kCUDA, static_cast< c10::DeviceIndex >(Device_Id)}).dtype(AT_kTYPE).requires_grad(true));
+
 #else
   keops_error(Error_msg_no_cuda);
   throw std::runtime_error("Simply here to avoid a warning at compilation.");
@@ -96,11 +97,10 @@ m.doc() = "pyKeOps: KeOps for pytorch through pybind11 (pytorch flavour).";
 
 m.def("genred_pytorch", &generic_red <at::Tensor, at::Tensor>, "Entry point to keops - pytorch version.");
 
-m.attr("tagIJ") = keops::TAGIJ;
-m.attr("dimout") = keops::DIMOUT;
-m.attr("formula") = keops::f;
-m.attr("compiled_formula") = xstr(keops::FORMULA_OBJ_STR);
-m.attr("compiled_aliases") = xstr(keops::VAR_ALIASES_STR);
+m.attr("tagIJ") = keops_binders::keops_tagIJ;
+m.attr("dimout") = keops_binders::keops_dimout;
+m.attr("formula") = keops_binders::keops_formula_string;
+
 }
 
 
