@@ -4,30 +4,50 @@ from pykeops.common.get_options import get_tag_backend
 from pykeops.common.keops_io import LoadKeOps
 from pykeops.common.operations import preprocess, postprocess
 from pykeops.torch.half2_convert import preprocess_half2, postprocess_half2
-from pykeops.common.parse_type import get_type, get_sizes, complete_aliases
-from pykeops.common.parse_type import get_optional_flags
+from pykeops.common.parse_type import (
+    get_type,
+    get_sizes,
+    complete_aliases,
+    get_optional_flags,
+)
 from pykeops.common.utils import axis2cat
 from pykeops.torch import default_dtype, include_dirs
 
+
 class GenredAutograd(torch.autograd.Function):
     """
     This class is the entry point to pytorch auto grad engine.
     """
 
     @staticmethod
-    def forward(ctx, formula, aliases, backend, dtype, device_id, ranges, optional_flags, rec_multVar_highdim, *args):
-        
+    def forward(
+        ctx,
+        formula,
+        aliases,
+        backend,
+        dtype,
+        device_id,
+        ranges,
+        optional_flags,
+        rec_multVar_highdim,
+        nx,
+        ny,
+        *args
+    ):
+
         # N.B. when rec_multVar_highdim option is set, it means that formula is of the form "sum(F*b)", where b is a variable
         # with large dimension. In this case we set compiler option MULT_VAR_HIGHDIM to allow for the use of the special "final chunk" computation
         # mode. However, this may not be also true for the gradients of the same formula. In fact only the gradient
-        # with respect to variable b will have the same form. Hence, we save optional_flags current status into ctx, 
+        # with respect to variable b will have the same form. Hence, we save optional_flags current status into ctx,
         # before adding the MULT_VAR_HIGHDIM compiler option.
         ctx.optional_flags = optional_flags.copy()
         if rec_multVar_highdim is not None:
-            optional_flags += ['-DMULT_VAR_HIGHDIM=1']
+            optional_flags += ["-DMULT_VAR_HIGHDIM=1"]
+
+        myconv = LoadKeOps(
+            formula, aliases, dtype, "torch", optional_flags, include_dirs
+        ).import_module()
 
-        myconv = LoadKeOps(formula, aliases, dtype, 'torch', optional_flags+include_dirs).import_module()
-        
         # Context variables: save everything to compute the gradient:
         ctx.formula = formula
         ctx.aliases = aliases
@@ -37,21 +57,28 @@ class GenredAutograd(torch.autograd.Function):
         ctx.ranges = ranges
         ctx.rec_multVar_highdim = rec_multVar_highdim
         ctx.myconv = myconv
+        ctx.nx = nx
+        ctx.ny = ny
 
         tagCPUGPU, tag1D2D, tagHostDevice = get_tag_backend(backend, args)
 
-        if tagCPUGPU==1 & tagHostDevice==1:
+        if tagCPUGPU == 1 & tagHostDevice == 1:
             device_id = args[0].device.index
-            for i in range(1,len(args)):
+            for i in range(1, len(args)):
                 if args[i].device.index != device_id:
-                    raise ValueError("[KeOps] Input arrays must be all located on the same device.")
-        
-        if ranges is None : ranges = () # To keep the same type
+                    raise ValueError(
+                        "[KeOps] Input arrays must be all located on the same device."
+                    )
+
+        if ranges is None:
+            ranges = ()  # To keep the same type
 
         # N.B.: KeOps C++ expects contiguous integer arrays as ranges
         ranges = tuple(r.contiguous() for r in ranges)
 
-        result = myconv.genred_pytorch(tagCPUGPU, tag1D2D, tagHostDevice, device_id, ranges, *args)
+        result = myconv.genred_pytorch(
+            tagCPUGPU, tag1D2D, tagHostDevice, device_id, ranges, nx, ny, *args
+        )
 
         # relying on the 'ctx.saved_variables' attribute is necessary  if you want to be able to differentiate the output
         #  of the backward once again. It helps pytorch to keep track of 'who is who'.
@@ -69,78 +96,154 @@ class GenredAutograd(torch.autograd.Function):
         optional_flags = ctx.optional_flags
         device_id = ctx.device_id
         myconv = ctx.myconv
+        nx = ctx.nx
+        ny = ctx.ny
         args = ctx.saved_tensors[:-1]  # Unwrap the saved variables
         nargs = len(args)
         result = ctx.saved_tensors[-1].detach()
 
-        not_supported = ["Min_ArgMin_Reduction", "Min_Reduction",
-                         "Max_ArgMax_Reduction", "Max_Reduction",
-                         "KMin_ArgKMin_Reduction", "KMin_Reduction"]
+        not_supported = [
+            "Min_ArgMin_Reduction",
+            "Min_Reduction",
+            "Max_ArgMax_Reduction",
+            "Max_Reduction",
+            "KMin_ArgKMin_Reduction",
+            "KMin_Reduction",
+        ]
         for red in not_supported:
             if formula.startswith(red):
-                raise NotImplementedError("As of today, KeOps does not support "
-                                          + "backpropagation through the " + red + " reduction. "
-                                          + "Adding this feature to LazyTensors is on the cards "
-                                          + "for future releases... But until then, you may want "
-                                          + "to consider extracting the relevant integer indices "
-                                          + "with a '.argmin()', '.argmax()' or '.argKmin()' reduction "
-                                          + "before using PyTorch advanced indexing to create a fully-differentiable "
-                                          + "tensor containing the relevant 'minimal' values.")
+                raise NotImplementedError(
+                    "As of today, KeOps does not support "
+                    + "backpropagation through the "
+                    + red
+                    + " reduction. "
+                    + "Adding this feature to LazyTensors is on the cards "
+                    + "for future releases... But until then, you may want "
+                    + "to consider extracting the relevant integer indices "
+                    + "with a '.argmin()', '.argmax()' or '.argKmin()' reduction "
+                    + "before using PyTorch advanced indexing to create a fully-differentiable "
+                    + "tensor containing the relevant 'minimal' values."
+                )
 
         # If formula takes 5 variables (numbered from 0 to 4), then the gradient
         # wrt. the output, G, should be given as a 6-th variable (numbered 5),
         # with the same dim-cat as the formula's output.
-        eta = 'Var(' + str(nargs) + ',' + str(myconv.dimout) + ',' + str(myconv.tagIJ) + ')'
+        eta = (
+            "Var("
+            + str(nargs)
+            + ","
+            + str(myconv.dimout)
+            + ","
+            + str(myconv.tagIJ)
+            + ")"
+        )
 
         # there is also a new variable for the formula's output
-        resvar = 'Var(' + str(nargs+1) + ',' + str(myconv.dimout) + ',' + str(myconv.tagIJ) + ')'
-        
+        resvar = (
+            "Var("
+            + str(nargs + 1)
+            + ","
+            + str(myconv.dimout)
+            + ","
+            + str(myconv.tagIJ)
+            + ")"
+        )
+
         grads = []  # list of gradients wrt. args;
 
-        for (var_ind, (sig, arg_ind)) in enumerate(zip(aliases, args)):  # Run through the arguments
+        for (var_ind, (sig, arg_ind)) in enumerate(
+            zip(aliases, args)
+        ):  # Run through the arguments
             # If the current gradient is to be discarded immediatly...
-            if not ctx.needs_input_grad[var_ind + 8]:  # because of (formula, aliases, backend, dtype, device_id, ranges, optional_flags, rec_multVar_highdim)
+            if not ctx.needs_input_grad[
+                var_ind + 10
+            ]:  # because of (formula, aliases, backend, dtype, device_id, ranges, optional_flags, rec_multVar_highdim, nx, ny)
                 grads.append(None)  # Don't waste time computing it.
 
-            else:  
+            else:
                 # Otherwise, the current gradient is really needed by the user. Adding new aliases is way too dangerous if we want to compute
                 # second derivatives, etc. So we make explicit references to Var<ind,dim,cat> instead.
                 # New here (Joan) : we still add the new variables to the list of "aliases" (without
                 # giving new aliases for them) these will not be used in the C++ code,
                 # but are useful to keep track of the actual variables used in the formula
                 _, cat, dim, pos = get_type(sig, position_in_list=var_ind)
-                var = 'Var(' + str(pos) + ',' + str(dim) + ',' + str(cat) + ')'  # V
-                formula_g = 'Grad_WithSavedForward(' + formula + ', ' + var + ', ' + eta + ', ' + resvar + ')'  # Grad<F,V,G,R>
+                var = "Var(" + str(pos) + "," + str(dim) + "," + str(cat) + ")"  # V
+                formula_g = (
+                    "Grad_WithSavedForward("
+                    + formula
+                    + ", "
+                    + var
+                    + ", "
+                    + eta
+                    + ", "
+                    + resvar
+                    + ")"
+                )  # Grad<F,V,G,R>
                 aliases_g = aliases + [eta, resvar]
-                args_g = args + (G,) + (result,)  # Don't forget the gradient to backprop !
-                
+                args_g = (
+                    args + (G,) + (result,)
+                )  # Don't forget the gradient to backprop !
+
                 # N.B.: if I understand PyTorch's doc, we should redefine this function every time we use it?
                 genconv = GenredAutograd().apply
 
                 # For a reduction of the type sum(F*b), with b a variable, and if we require the gradient
                 # with respect to b, the gradient will be of same type sum(F*eta). So we set again rec_multVar option
                 # in this case.
-                if pos==ctx.rec_multVar_highdim:
-                    rec_multVar_highdim = nargs # nargs is the position of variable eta.
+                if pos == ctx.rec_multVar_highdim:
+                    rec_multVar_highdim = (
+                        nargs  # nargs is the position of variable eta.
+                    )
                 else:
                     rec_multVar_highdim = None
 
-                if cat == 2:  # we're referring to a parameter, so we'll have to sum both wrt 'i' and 'j'
+                if (
+                    cat == 2
+                ):  # we're referring to a parameter, so we'll have to sum both wrt 'i' and 'j'
                     # WARNING !! : here we rely on the implementation of DiffT in files in folder keops/core/formulas/reductions
                     # if tagI==cat of V is 2, then reduction is done wrt j, so we need to further sum output wrt i
-                    grad = genconv(formula_g, aliases_g, backend, dtype, device_id, ranges, optional_flags, rec_multVar_highdim, *args_g)
+                    grad = genconv(
+                        formula_g,
+                        aliases_g,
+                        backend,
+                        dtype,
+                        device_id,
+                        ranges,
+                        optional_flags,
+                        rec_multVar_highdim,
+                        nx,
+                        ny,
+                        *args_g
+                    )
                     # Then, sum 'grad' wrt 'i' :
                     # I think that '.sum''s backward introduces non-contiguous arrays,
                     # and is thus non-compatible with GenredAutograd: grad = grad.sum(0)
                     # We replace it with a 'handmade hack' :
                     # grad = torch.ones(1, grad.shape[0]).type_as(grad.data) @ grad
                     # grad = grad.view(-1)
-                    grad = (1. * grad).sum(-2)
+                    grad = (1.0 * grad).sum(-2)
                     dims_to_collapse = tuple(
-                        i for (i, (x, y)) in enumerate(zip(arg_ind.shape[:-1], grad.shape[:-1])) if x < y)
+                        i
+                        for (i, (x, y)) in enumerate(
+                            zip(arg_ind.shape[:-1], grad.shape[:-1])
+                        )
+                        if x < y
+                    )
 
                 else:
-                    grad = genconv(formula_g, aliases_g, backend, dtype, device_id, ranges, optional_flags, rec_multVar_highdim, *args_g)
+                    grad = genconv(
+                        formula_g,
+                        aliases_g,
+                        backend,
+                        dtype,
+                        device_id,
+                        ranges,
+                        optional_flags,
+                        rec_multVar_highdim,
+                        nx,
+                        ny,
+                        *args_g
+                    )
 
                     # N.B.: 'grad' is always a full [A, .., B, M, D] or [A, .., B, N, D] or [A, .., B, D] tensor,
                     #       whereas 'arg_ind' may have some broadcasted batched dimensions.
@@ -148,56 +251,77 @@ class GenredAutograd(torch.autograd.Function):
                     #       which is the adjoint of the good old "repmat" that could have been used
                     #       to emulate the batch broadcasting.
                     dims_to_collapse = tuple(
-                        i for (i, (x, y)) in enumerate(zip(arg_ind.shape[:-2], grad.shape[:-2])) if x < y)
+                        i
+                        for (i, (x, y)) in enumerate(
+                            zip(arg_ind.shape[:-2], grad.shape[:-2])
+                        )
+                        if x < y
+                    )
 
                 if dims_to_collapse != ():
-                    grad = (1. * grad).sum(dims_to_collapse, keepdim=True)
-                grad = grad.reshape(arg_ind.shape)  # The gradient should have the same shape as the input!
+                    grad = (1.0 * grad).sum(dims_to_collapse, keepdim=True)
+                grad = grad.reshape(
+                    arg_ind.shape
+                )  # The gradient should have the same shape as the input!
                 grads.append(grad)
-        
-        # Grads wrt. formula, aliases, backend, dtype, device_id, ranges, optional_flags, rec_multVar_highdim, *args
-        return (None, None, None, None, None, None, None, None, *grads)
+
+        # Grads wrt. formula, aliases, backend, dtype, device_id, ranges, optional_flags, rec_multVar_highdim, nx, ny, *args
+        return (None, None, None, None, None, None, None, None, None, None, *grads)
 
 
-class Genred():
+class Genred:
     r"""
-        Creates a new generic operation.
+    Creates a new generic operation.
 
-        This is KeOps' main function, whose usage is documented in
-        the :doc:`user-guide <../../Genred>`,
-        the :doc:`gallery of examples <../../../_auto_examples/index>`
-        and the :doc:`high-level tutorials <../../../_auto_tutorials/index>`.
-        Taking as input a handful of strings and integers that specify
-        a custom Map-Reduce operation, it returns a C++ wrapper
-        that can be called just like any other PyTorch function.
+    This is KeOps' main function, whose usage is documented in
+    the :doc:`user-guide <../../Genred>`,
+    the :doc:`gallery of examples <../../../_auto_examples/index>`
+    and the :doc:`high-level tutorials <../../../_auto_tutorials/index>`.
+    Taking as input a handful of strings and integers that specify
+    a custom Map-Reduce operation, it returns a C++ wrapper
+    that can be called just like any other PyTorch function.
 
-        Note:
-            :func:`Genred` is fully compatible with PyTorch's :mod:`autograd` engine:
-            You can **backprop** through the KeOps :meth:`__call__` just
-            as if it was a vanilla PyTorch operation (except for Min or Max reduction types,
-            see :ref:`reductions <part.reduction>`)
+    Note:
+        :func:`Genred` is fully compatible with PyTorch's :mod:`autograd` engine:
+        You can **backprop** through the KeOps :meth:`__call__` just
+        as if it was a vanilla PyTorch operation (except for Min or Max reduction types,
+        see :ref:`reductions <part.reduction>`)
 
-        Example:
-            >>> my_conv = Genred('Exp(-SqNorm2(x - y))',  # formula
-            ...                  ['x = Vi(3)',            # 1st input: dim-3 vector per line
-            ...                   'y = Vj(3)'],           # 2nd input: dim-3 vector per column
-            ...                  reduction_op='Sum',      # we also support LogSumExp, Min, etc.
-            ...                  axis=1)                  # reduce along the lines of the kernel matrix
-            >>> # Apply it to 2d arrays x and y with 3 columns and a (huge) number of lines
-            >>> x = torch.randn(1000000, 3, requires_grad=True).cuda()
-            >>> y = torch.randn(2000000, 3).cuda()
-            >>> a = my_conv(x, y)  # a_i = sum_j exp(-|x_i-y_j|^2)
-            >>> print(a.shape)
-            torch.Size([1000000, 1])
-            >>> [g_x] = torch.autograd.grad((a ** 2).sum(), [x])  # KeOps supports autograd!
-            >>> print(g_x.shape)
-            torch.Size([1000000, 3])
+    Example:
+        >>> my_conv = Genred('Exp(-SqNorm2(x - y))',  # formula
+        ...                  ['x = Vi(3)',            # 1st input: dim-3 vector per line
+        ...                   'y = Vj(3)'],           # 2nd input: dim-3 vector per column
+        ...                  reduction_op='Sum',      # we also support LogSumExp, Min, etc.
+        ...                  axis=1)                  # reduce along the lines of the kernel matrix
+        >>> # Apply it to 2d arrays x and y with 3 columns and a (huge) number of lines
+        >>> x = torch.randn(1000000, 3, requires_grad=True).cuda()
+        >>> y = torch.randn(2000000, 3).cuda()
+        >>> a = my_conv(x, y)  # a_i = sum_j exp(-|x_i-y_j|^2)
+        >>> print(a.shape)
+        torch.Size([1000000, 1])
+        >>> [g_x] = torch.autograd.grad((a ** 2).sum(), [x])  # KeOps supports autograd!
+        >>> print(g_x.shape)
+        torch.Size([1000000, 3])
 
-        """
-    
-    def __init__(self, formula, aliases, reduction_op='Sum', axis=0, dtype=default_dtype, opt_arg=None,
-                 formula2=None, cuda_type=None, dtype_acc="auto", use_double_acc=False, sum_scheme="auto", 
-                 enable_chunks=True, optional_flags=[], rec_multVar_highdim=None):
+    """
+
+    def __init__(
+        self,
+        formula,
+        aliases,
+        reduction_op="Sum",
+        axis=0,
+        dtype=default_dtype,
+        opt_arg=None,
+        formula2=None,
+        cuda_type=None,
+        dtype_acc="auto",
+        use_double_acc=False,
+        sum_scheme="auto",
+        enable_chunks=True,
+        optional_flags=[],
+        rec_multVar_highdim=None,
+    ):
         r"""
         Instantiate a new generic operation.
 
@@ -247,54 +371,71 @@ class Genred():
             opt_arg (int, default = None): If **reduction_op** is in ``["KMin", "ArgKMin", "KMin_ArgKMin"]``,
                 this argument allows you to specify the number ``K`` of neighbors to consider.
 
-            dtype_acc (string, default ``"auto"``): type for accumulator of reduction, before casting to dtype. 
+            dtype_acc (string, default ``"auto"``): type for accumulator of reduction, before casting to dtype.
                 It improves the accuracy of results in case of large sized data, but is slower.
-                Default value "auto" will set this option to the value of dtype. The supported values are: 
+                Default value "auto" will set this option to the value of dtype. The supported values are:
 
                   - **dtype_acc** = ``"float16"`` : allowed only if dtype is "float16".
                   - **dtype_acc** = ``"float32"`` : allowed only if dtype is "float16" or "float32".
                   - **dtype_acc** = ``"float64"`` : allowed only if dtype is "float32" or "float64"..
 
             use_double_acc (bool, default False): same as setting dtype_acc="float64" (only one of the two options can be set)
-                If True, accumulate results of reduction in float64 variables, before casting to float32. 
+                If True, accumulate results of reduction in float64 variables, before casting to float32.
                 This can only be set to True when data is in float32 or float64.
                 It improves the accuracy of results in case of large sized data, but is slower.
-           
+
             sum_scheme (string, default ``"auto"``): method used to sum up results for reductions.
                 Default value "auto" will set this option to "block_red". Possible values are:
                   - **sum_scheme** =  ``"direct_sum"``: direct summation
-                  - **sum_scheme** =  ``"block_sum"``: use an intermediate accumulator in each block before accumulating 
-                    in the output. This improves accuracy for large sized data. 
+                  - **sum_scheme** =  ``"block_sum"``: use an intermediate accumulator in each block before accumulating
+                    in the output. This improves accuracy for large sized data.
                   - **sum_scheme** =  ``"kahan_scheme"``: use Kahan summation algorithm to compensate for round-off errors. This improves
-                accuracy for large sized data. 
-                
+                accuracy for large sized data.
+
             enable_chunks (bool, default True): enable automatic selection of special "chunked" computation mode for accelerating reductions
-				with formulas involving large dimension variables.
-				
-			optional_flags (list, default []): further optional flags passed to the compiler, in the form ['-D...=...','-D...=...']
+                                with formulas involving large dimension variables.
+
+                        optional_flags (list, default []): further optional flags passed to the compiler, in the form ['-D...=...','-D...=...']
 
         """
         if cuda_type:
             # cuda_type is just old keyword for dtype, so this is just a trick to keep backward compatibility
-            dtype = cuda_type 
+            dtype = cuda_type
         self.reduction_op = reduction_op
         reduction_op_internal, formula2 = preprocess(reduction_op, formula2)
-        
-        self.optional_flags = optional_flags + get_optional_flags(reduction_op_internal, dtype_acc, use_double_acc, sum_scheme, dtype, enable_chunks)
-        
-        str_opt_arg = ',' + str(opt_arg) if opt_arg else ''
-        str_formula2 = ',' + formula2 if formula2 else ''
-        
-        self.formula = reduction_op_internal + '_Reduction(' + formula + str_opt_arg + ',' + str(
-            axis2cat(axis)) + str_formula2 + ')'
-        self.aliases = complete_aliases(self.formula, list(aliases)) # just in case the user provided a tuple
+
+        self.optional_flags = optional_flags + get_optional_flags(
+            reduction_op_internal,
+            dtype_acc,
+            use_double_acc,
+            sum_scheme,
+            dtype,
+            enable_chunks,
+        )
+
+        str_opt_arg = "," + str(opt_arg) if opt_arg else ""
+        str_formula2 = "," + formula2 if formula2 else ""
+
+        self.formula = (
+            reduction_op_internal
+            + "_Reduction("
+            + formula
+            + str_opt_arg
+            + ","
+            + str(axis2cat(axis))
+            + str_formula2
+            + ")"
+        )
+        self.aliases = complete_aliases(
+            self.formula, list(aliases)
+        )  # just in case the user provided a tuple
         self.dtype = dtype
         self.axis = axis
         self.opt_arg = opt_arg
-        
+
         self.rec_multVar_highdim = rec_multVar_highdim
 
-    def __call__(self, *args, backend='auto', device_id=-1, ranges=None):
+    def __call__(self, *args, backend="auto", device_id=-1, ranges=None):
         r"""
         To apply the routine on arbitrary torch Tensors.
 
@@ -403,25 +544,44 @@ class Genred():
         """
 
         nx, ny = get_sizes(self.aliases, *args)
-        nout, nred = (nx, ny) if self.axis==1 else (ny, nx)
+        nout, nred = (nx, ny) if self.axis == 1 else (ny, nx)
 
         if "Arg" in self.reduction_op:
             # when using Arg type reductions,
             # if nred is greater than 16 millions and dtype=float32, the result is not reliable
             # because we encode indices as floats, so we raise an exception ;
             # same with float16 type and nred>2048
-            if nred>1.6e7 and self.dtype in ("float32","float"):
-                raise ValueError('size of input array is too large for Arg type reduction with single precision. Use double precision.')  
-            elif nred>2048 and self.dtype in ("float16","half"):
-                raise ValueError('size of input array is too large for Arg type reduction with float16 dtype..')  
+            if nred > 1.6e7 and self.dtype in ("float32", "float"):
+                raise ValueError(
+                    "size of input array is too large for Arg type reduction with single precision. Use double precision."
+                )
+            elif nred > 2048 and self.dtype in ("float16", "half"):
+                raise ValueError(
+                    "size of input array is too large for Arg type reduction with float16 dtype.."
+                )
 
-        if self.dtype in ('float16','half'):
-            args, ranges, tag_dummy, N = preprocess_half2(args, self.aliases, self.axis, ranges, nx, ny)
-        
-        out = GenredAutograd.apply(self.formula, self.aliases, backend, self.dtype, 
-                                   device_id, ranges, self.optional_flags, self.rec_multVar_highdim, *args)
+        if self.dtype in ("float16", "half"):
+            args, ranges, tag_dummy, N = preprocess_half2(
+                args, self.aliases, self.axis, ranges, nx, ny
+            )
 
-        if self.dtype in ('float16','half'):
+        out = GenredAutograd.apply(
+            self.formula,
+            self.aliases,
+            backend,
+            self.dtype,
+            device_id,
+            ranges,
+            self.optional_flags,
+            self.rec_multVar_highdim,
+            nx,
+            ny,
+            *args
+        )
+
+        if self.dtype in ("float16", "half"):
             out = postprocess_half2(out, tag_dummy, self.reduction_op, N)
 
-        return postprocess(out, "torch", self.reduction_op, nout, self.opt_arg, self.dtype)
+        return postprocess(
+            out, "torch", self.reduction_op, nout, self.opt_arg, self.dtype
+        )
